 #!/bin/usr/perl/

# This script takes a genotype file and various annotations files from either Affymetrix or Illumina.
# It produces four output files formated for alohomora_m: the genotype file, the genetic map file,
# and the allele frequency file.
# The program also produces a fourth output file which lists the correspondance between the AB and ACGT format.

use strict;
# use warnings;

use Getopt::Long;
use Tk;
use Tk::ROText;
use Tk::FileSelect;
use Tk::BrowseEntry;
use Tk::Balloon;

######                              ######  
### Variables and constant declaration ###
######                              ######

# Some constant
use constant TRUE   => 1;
use constant FALSE  => 0;

## Global variable declarations 
our $abort             = FALSE;
our $batch             = FALSE;
our $GUI               = TRUE;
our $chip              = "GenomeWideSNP_6";
our $chip_number       = 1;
our $input_geno        = undef;
our $input_anno        = undef;
our $input_map         = undef;
our $input_skip        = undef;
our $population        = "Caucasian";
our $population_number = 1;
our $input_matrix      = undef;
our $output_name       = undef;
our $format            = "AB";
our $affy_chips_ref    = [ 'GenomeWideSNP_6' ];   # to be implemented: 'Mapping250k_Nsp', 'Mapping250k_Sty', 'Mapping50k_Hind240', 'Mapping50k_Xba240',
our $illu_chips_ref    = [ 'Human610_Quad_v1_b']; # to be implemented: 'HumanHap_650Y_v3', 'HumanHap550_v3', 'HumanHap300_v1',
our ( $OUTPUT_GENO, $OUTPUT_MAP, $OUTPUT_FREQ, $OUTPUT_KEY, $TMP_WRITING);
our ( $INPUT_GENO, $INPUT_ANNO, $INPUT_MAP, $INPUT_MATRIX, $TMP_READING);
    
# Lists of founders within the Illumina population matrix files.
# These lists will be used for the calculation of the allele frequencies for a given population.
our $founders_list_CEU =  "NA12892 NA12891 NA12875 NA12874 NA12873 NA12872 NA12815 NA12814 NA12813 NA12812 NA12763 NA12762 NA12761 NA12760 NA12751 NA12750 NA12717 NA12716 NA12264 NA12249 NA12234 NA12155 NA12154 NA12146 NA12145 NA12144 NA12057 NA12044 NA12043 NA12006 NA12005 NA12004 NA12003 NA11995 NA11994 NA11993 NA11992 NA11882 NA11881 NA11840 NA11832 NA11830 NA11829 NA07357 NA07056 NA07034 NA07022 NA07000 NA06994 NA06993 NA06985";
our $founders_list_YRI =  "NA18501 NA18502 NA18505 NA18507 NA18508 NA18517 NA18852 NA18853 NA18856 NA18859 NA18861 NA18913 NA19099 NA19101 NA19102 NA19116 NA19128 NA19138 NA19140 NA19141 NA19143 NA19152 NA19160 NA19171 NA19172 NA19192 NA19201 NA19203 NA19204 NA19206 NA19207 NA19210 NA19222 NA19239 NA19209 NA19200 NA19193 NA19153 NA19144 NA19119 NA19093 NA19092 NA18912 NA18871 NA18870 NA18862 NA18855 NA18516 NA18504";
our $founders_list_CHB =  "NA18637 NA18635 NA18633 NA18632 NA18624 NA18623 NA18622 NA18621 NA18620 NA18612 NA18611 NA18609 NA18608 NA18605 NA18603 NA18594 NA18593 NA18592 NA18582 NA18577 NA18573 NA18572 NA18571 NA18570 NA18566 NA18564 NA18563 NA18562 NA18561 NA18552 NA18550 NA18547 NA18545 NA18542 NA18537 NA18532 NA18529 NA18524";
our $founders_list_JPT =  "NA19012 NA19007 NA19005 NA19003 NA18999 NA18997 NA18992 NA18991 NA18990 NA18987 NA18981 NA18980 NA18978 NA18975 NA18974 NA18973 NA18972 NA18971 NA18969 NA18968 NA18967 NA18966 NA18965 NA18964 NA18961 NA18960 NA18959 NA18956 NA18953 NA18952 NA18951 NA18948 NA18947 NA18945 NA18944 NA18943 NA18942";

# Local variable declarations
my $version            = "1.0.2";
my $last_update        = "September 29th, 2009";
my $provider           = "Affymetrix";
my $user_input         = undef;
my $user_sub_input     = undef;


our $time_start_compute = 0;
our $time_start_sort = 0;
our $time_end_compute = 0;
our $time_end_sort = 0;
our $sort_counter = 0;


######        ###### 
###   Batch-mode ###
######        ######

### Get option from the command line (if any) :
GetOptions ( 'batch'        =>  \$batch,
	     'provider=s'   =>  \$provider,
	     'chip=i'       =>  \$chip_number,
	     'genotype=s'   =>  \$input_geno,
	     'annotation=s' =>  \$input_anno,
	     'output=s'     =>  \$output_name,
	     'map=s'        =>  \$input_map,
	     'matrix=s'     =>  \$input_matrix,
	     'population=i' =>  \$population_number,
	     'format=s'     =>  \$format, 
	     'skiplist=s'   =>  \$input_skip,
	   );

# Creates a flag for the GUI.
if ( $batch ) { $GUI = FALSE; }

# Handle the batch mode
if ( $batch ) {

    ##### Assign command line options:
    
    # Set the $population variable according to $population_number
    if ( $population_number == 1 ) {
	$population = "Caucasian";}
    elsif ( $population_number == 2 ) {
	$population = "Han Chinese";}
    elsif ( $population == 3) {
	$population = "Japanese";}
    elsif ( $population_number == 4 ) {
	$population = "Yoruban";}
    
    
    ##### Verify the validity of input files
    
    # Check the validity of the provider choice
    if ( $provider ne "Affymetrix" && $provider ne "Illumina" ) {
	die "The option \"--provider=$provider\" was not recognized.\n"
	   ."Please use \"--provider=Affymetrix\" or \"--provider=Illumina\".\n";}
	
    # Check the validity of the format choice
    if ( $format ne "AB" && $format ne "ACGT-TOP" && $format ne "ACGT-FWD" ) {
	die "The option --format $format was not recognized.\n"
	   ."Please use \'--format=AB\' or \'--format=ACGT-TOP\' or \'--format=ACGT-FWD\'.\n";} 

    # Check the validity of the selected annotation file
    if (! $input_anno) {
	die "You have to indicate the name of the annotation file.\n";}
    if (&CheckReadMode($input_anno)) {
	die "Invalid annotation file.\n";}

    # Check the validity of the selected genotype file
    if (! $input_geno) {
	die "You have to indicate the name of the genotype file.\n";}
    if ( &CheckReadMode($input_geno) ) {
	die "Invalid genotype file.\n";}

    # Check the validity of the selected output name
    if ( &CheckWriteMode($output_name) ) {
	die "Invalid output name.\n";}
    
    if ( $provider eq "Illumina" ) {
	
	# Check the validity of the selected population matrix file
	if (! $input_matrix) {
	    die "You have to indicate the name of the matrix file.\n";}
	if (&CheckReadMode($input_matrix)) {
	    die "Invalid population matrix file.\n";}

	# Check the validity of the selected map file
	if (! $input_map) {
	    die "You have to indicate the name of the map file.\n";}
	if(&CheckReadMode($input_map)) {
	    die "Invalid map file.\n";}
    }

    # Launch the computation
    &Launch_formatting();
}


######                                        ######
###    Graphical User Interface initialization   ###
######                                        ######

# MAIN WINDOW
my $mw = MainWindow -> new(-title => "Chip2Spell $version\tPrepare raw data for the use of Alohomora", );
$mw -> geometry("850x650");

# FRAMES
my $provider_frame   = $mw -> Frame( -borderwidth => 3, -relief => "ridge", );
my $chip_frame       = $mw -> Frame( -borderwidth => 3, -relief => "ridge", );   
my $population_frame = $mw -> Frame( -borderwidth => 3, -relief => "ridge", );
my $file_frame       = $mw -> Frame( -borderwidth => 3, -relief => "ridge", );
my $illu_frame       = $mw -> Frame( -borderwidth => 3, -relief => "ridge", );

$provider_frame   -> place( -relx => 0.01, -relwidth => 0.32, -rely => 0.02, -relheight => 0.22 );
$chip_frame       -> place( -relx => 0.34, -relwidth => 0.32, -rely => 0.02, -relheight => 0.22 );
$population_frame -> place( -relx => 0.67, -relwidth => 0.32, -rely => 0.02, -relheight => 0.22 );
$file_frame       -> place( -relx => 0.01, -relwidth => 0.32, -rely => 0.26, -relheight => 0.31 );
$illu_frame       -> place( -relx => 0.01, -relwidth => 0.32, -rely => 0.59, -relheight => 0.31 ); 

# LOG 
our $log          = $mw -> Scrolled( "ROText", -scrollbars => "e", -borderwidth => 3, -relief => "ridge", -wrap => "word", );
$log              -> place( -relx => 0.34, -relwidth => 0.65, -rely => 0.26, -relheight => 0.31 ); 
$log              -> insert( "end", "LOG\nThis window will inform you of the progress of the computation.\n\n" );


# Integrity
our $integrity    = $mw -> Scrolled( "ROText", -scrollbars => "e", -borderwidth => 3, -relief => "ridge", -wrap => "word", );
$integrity        -> place( -relx => 0.34, -relwidth => 0.65, -rely => 0.59, -relheight => 0.31 ); 
$integrity        -> insert( "end", "INTEGRITY CHECK\nThis window will help you determine if Chip2Spell reads your input files properly. ".
			     "It will display information about what Chip2Spell reads in your file and what it expects to read.\n\n" );

# CHIP_FRAME
my $chip_label  = $chip_frame -> Label (-text => "Choose chip",   -background => "light blue", );
our $chip_be     = $chip_frame -> BrowseEntry( -label => "Chip: ", -variable => \$chip, -choices => $affy_chips_ref, -state => "readonly", );


# PROVIDER_FRAME
my $provider_label = $provider_frame -> Label( -text => "Choose provider", -background => "light blue", );
my $affy_rb = $provider_frame -> Radiobutton(  -text => "Affymetrix", -value => "Affymetrix", -variable => \$provider, -command => \&Affy_provider );
my $illu_rb = $provider_frame -> Radiobutton(  -text => "Illumina",   -value => "Illumina",   -variable => \$provider, -command => \&Illu_provider );

$provider_label -> place( -relx => 0.00, -relwidth => 1.00, -rely => 0.00, -relheight => 0.32 ); 
$affy_rb        -> place( -relx => 0.01, -relwidth => 0.48, -rely => 0.34, -relheight => 0.31 ); 
$illu_rb        -> place( -relx => 0.51, -relwidth => 0.48, -rely => 0.34, -relheight => 0.31 ); 


$chip_label  -> place( -relx => 0.00, -relwidth => 1.00, -rely => 0.00, -relheight => 0.32 ); 
$chip_be     -> place( -relx => 0.01, -relwidth => 0.98, -rely => 0.34, -relheight => 0.31 ); 


# POPULATION_FRAME
my $population_label = $population_frame -> Label ( -text => "Choose population", -background => "light blue",  );
our $ceu_rb = $population_frame -> Radiobutton( -text => "Caucasian",   -value => "Caucasian",   -variable => \$population, -anchor => "w", );
our $hcn_rb = $population_frame -> Radiobutton( -text => "Han Chinese", -value => "Han Chinese", -variable => \$population, -anchor => "w", );
our $jpt_rb = $population_frame -> Radiobutton( -text => "Japanese",    -value => "Japanese",    -variable => \$population, -anchor => "w", ); 
our $yrb_rb = $population_frame -> Radiobutton( -text => "Yoruban",     -value => "Yoruban",     -variable => \$population, -anchor => "w", );

$population_label -> place( -relx => 0.00, -relwidth => 1.00, -rely => 0.00, -relheight => 0.32 ); 
$ceu_rb           -> place( -relx => 0.01, -relwidth => 0.48, -rely => 0.34, -relheight => 0.31 );  
$hcn_rb           -> place( -relx => 0.51, -relwidth => 0.48, -rely => 0.34, -relheight => 0.31 );  
$jpt_rb           -> place( -relx => 0.01, -relwidth => 0.48, -rely => 0.67, -relheight => 0.31 );  
$yrb_rb           -> place( -relx => 0.51, -relwidth => 0.48, -rely => 0.67, -relheight => 0.31 );  

# FILE_FRAME
my $file_label       = $file_frame -> Label ( -text => "Choose files", -background => "light blue", );
my $geno_entry       = $file_frame -> Entry ( -textvariable => \$input_geno,  -state => "readonly", );
my $anno_entry       = $file_frame -> Entry ( -textvariable => \$input_anno,  -state => "readonly", );
my $skip_list_entry  = $file_frame -> Entry ( -textvariable => \$input_skip,  -state => "readonly", );
my $output_entry     = $file_frame -> Entry ( -textvariable => \$output_name, -state => "readonly", ); 

our $geno_button      = $file_frame -> Button( 
                                              -text    => "Genotype file", 
                                              -command => sub {$input_geno = $mw -> getOpenFile( -title => "Genotype file",);}, 
                                            );
our $anno_button      = $file_frame -> Button( 
                                              -text    => "Annotation file", 
                                              -command => sub {$input_anno = $mw -> getOpenFile ( -title => "Annotation file",);},
											  );
our $skip_list_button = $file_frame -> Button( 
                                              -text    => "Skip list", 
					      -state   => "disabled", 
                                              -command => sub {$input_skip = $mw -> getOpenFile ( -title => "Skip list", );},
                                             );
our $output_button    = $file_frame -> Button( 
                                              -text => "Output name",
					      -command => sub {$output_name = $mw -> getSaveFile( -title => "Output name", );},
					     );


$file_label       -> place( -relx => 0.00, -relwidth => 1.00, -rely => 0.00, -relheight => 0.22, );
$geno_button      -> place( -relx => 0.01, -relwidth => 0.35, -rely => 0.23, -relheight => 0.15, );
$anno_button      -> place( -relx => 0.01, -relwidth => 0.35, -rely => 0.39, -relheight => 0.15, );
$skip_list_button -> place( -relx => 0.01, -relwidth => 0.35, -rely => 0.55, -relheight => 0.15, );
$output_button    -> place( -relx => 0.01, -relwidth => 0.35, -rely => 0.71, -relheight => 0.15, );
$geno_entry       -> place( -relx => 0.37, -relwidth => 0.62, -rely => 0.23, -relheight => 0.15, );
$anno_entry       -> place( -relx => 0.37, -relwidth => 0.62, -rely => 0.39, -relheight => 0.15, );
$skip_list_entry  -> place( -relx => 0.37, -relwidth => 0.62, -rely => 0.55, -relheight => 0.15, );
$output_entry     -> place( -relx => 0.37, -relwidth => 0.62, -rely => 0.71, -relheight => 0.15, );

# ILLU_FRAME
my $illu_label      = $illu_frame -> Label(  -text => "Illumina options", -background => "light blue", );
my $illu_file_label = $illu_frame -> Label(  -text => "Choose files", );
my $map_entry       = $illu_frame -> Entry(  -textvariable => \$input_map,    -state => "readonly", );
my $matrix_entry    = $illu_frame -> Entry(  -textvariable => \$input_matrix, -state => "readonly", );
my $format_label    = $illu_frame -> Label(  -text => "Choose format", );
our $ab_rb           = $illu_frame -> Radiobutton( -text => "AB",       -value => "AB",       -variable => \$format, -state => "disabled" );
our $acgt_top_rb     = $illu_frame -> Radiobutton( -text => "ACGT-top", -value => "ACGT-TOP", -variable => \$format, -state => "disabled" );
our $acgt_fwd_rb     = $illu_frame -> Radiobutton( -text => "ACGT-fwd", -value => "ACGT-FWD", -variable => \$format, -state => "disabled" );

our $map_button      = $illu_frame -> Button( 
                                            -text    => "Map file", 
					    -state   => "disabled",
                                            -command => sub {$input_map = $mw -> getOpenFile( -title => "Map file", );},
                                           );
our $matrix_button   = $illu_frame -> Button( 
                                            -text    => "Matrix file",
					    -state   => "disabled",
                                            -command => sub {$input_matrix = $mw -> getOpenFile( -title => "Matrix file",);}, 
                                           );

$illu_label      -> place( -relx => 0.00, -relwidth => 1.00, -rely => 0.01, -relheight => 0.22, );
$illu_file_label -> place( -relx => 0.00, -relwidth => 1.00, -rely => 0.23, -relheight => 0.14, );
$map_button      -> place( -relx => 0.01, -relwidth => 0.35, -rely => 0.36, -relheight => 0.14, );
$matrix_button   -> place( -relx => 0.01, -relwidth => 0.35, -rely => 0.51, -relheight => 0.14, );
$map_entry       -> place( -relx => 0.37, -relwidth => 0.62, -rely => 0.36, -relheight => 0.14, );
$matrix_entry    -> place( -relx => 0.37, -relwidth => 0.62, -rely => 0.51, -relheight => 0.14, );
$format_label    -> place( -relx => 0.00, -relwidth => 1.00, -rely => 0.66, -relheight => 0.14, );
$acgt_top_rb     -> place( -relx => 0.01, -relwidth => 0.32, -rely => 0.81, -relheight => 0.14, );
$acgt_fwd_rb     -> place( -relx => 0.34, -relwidth => 0.32, -rely => 0.81, -relheight => 0.14, );
$ab_rb           -> place( -relx => 0.67, -relwidth => 0.32, -rely => 0.81, -relheight => 0.14, );

# BOTTOM BUTTONS
my  $reset_button = $mw -> Button( -text => "Reset", -background => "red",   -command => \&Reset,             );
our $start_button = $mw -> Button( -text => "Start", -background => "green", -command => \&Launch_formatting_GUI, );
my  $exit_button  = $mw -> Button( -text => "Exit" , -background => "red",   -command => sub {exit;},         );

$reset_button -> place( -relx => 0.17, -relwidth => 0.16, -rely => 0.92, -relheight => 0.07, );
$start_button -> place( -relx => 0.34, -relwidth => 0.32, -rely => 0.92, -relheight => 0.07, );
$exit_button  -> place( -relx => 0.67, -relwidth => 0.16, -rely => 0.92, -relheight => 0.07, );

# HELP BALLOON
my $help_balloon = $mw -> Balloon(
				  -state => 'balloon',
				  -initwait => 500,
				  );

$help_balloon -> attach( $geno_button,
			 -msg => "This file should contain your actual genotyping data.\n".
			         "See the documentation to know the detail of what Chip2Spell expects.",
			 );
$help_balloon -> attach( $anno_button,
			 -msg => "This file should be the latest annotation file available for the genotyping chip used.\n".
			         "You can download Illumina annotation files at \"www.illumina.com/ftp.ilmn\".\n".
			         "You can download Affymetrix annotation files at \"http://www.affymetrix.com/support/technical/annotationfilesmain.affx\".\n".
			         "See the documentation to know the detail of what Chip2Spell expects.",
			 );
$help_balloon -> attach( $skip_list_button,
			 -msg => "This functionnality is not currently implemented.\n".
			         "In the future, this will allow you to provide a list of SNPs that will\n".
			         "be excluded when your output files are generated.",
			 );

$help_balloon -> attach( $output_button,
			 -msg => "The output name that you provide will be used to produce four output files\n".
			         "by attaching the extensions \".geno\", \".map\", \".freq\" and \".key\"/n".
			         "See the documentation for details on these files.",
			 );

$help_balloon -> attach( $map_button,
			 -msg => "The \'map file\' is a documentation file from Illumina.\n".
			         "You can download it at \"www.illumina.com/ftp.ilmn\".\n".
			         "See the documentation for the detail of what Chip2Spell expects.",
			 );

$help_balloon -> attach( $matrix_button,
			 -msg => "The \'matrix file\' is a full genotype report in matrix format for a given population.\n".
			         "This genotype report is provided as documentation by Illumina.\n".
			         "You can download it at \"www.illumina.com/ftp.ilmn\".\n".
			         "This file will be used to calculate allele frequencies for a population.\n".
			         "See the documentation (perldoc Chip2Spell.pl) for the detail of what Chip2Spell expects.",
			 );

$help_balloon -> attach( $ab_rb,
			 -msg => "The \"AB\" format is one of the three formats used by the Affymetrix plateform\n".
			         "to generate a Genotype_Report. Beware that there is no relation between the A and B alleles\n".
			         "and the major or minor allele.",
			 );

$help_balloon -> attach( $acgt_fwd_rb,
			 -msg => "The \"ACGT-forward\" format is one of the three formats used by the Illumina plateform\n".
			          "to generate a Genotype_Report. For each individual and each SNP, it gives the allele on the forward strand.",
			 );

$help_balloon -> attach( $acgt_top_rb,
			 -msg => "The \"ACGT-top\" format is one of the three formats used by the Illumina\n".
			         "plateform to generate a Genotype_Report. For each individual and each SNP, it gives the\n".
			         "allele on the \'Top\' strand. Beware that the \'Top\' strand is a choice made by Illumina.\n".
			         "There is no correspondance between Illumina\'s Top strand and the forward strand reported\n".
                                 "in major databases like dbSNP.",
			 );

### User interactions
if ( $GUI ) {
    MainLoop();
}
	

################################################################
###                      SUB-ROUTINES                        ###
################################################################

sub Reset {
    $provider     = "Affymetrix";
    $chip_be      -> configure( -choices => $affy_chips_ref, );	
    $format       = "AB";
    $population   = "Caucasian";
    $output_name  = undef;
    $input_geno   = undef;
    $input_anno   = undef;
    $input_skip   = undef;
    $input_map    = undef;
    $input_matrix = undef;
}

sub Launch_formatting_GUI {
    
    # Test the validity of input and produce error messages:
    if (! $input_geno ) {
	$mw -> messageBox ( -title   => "Input file missing",
			    -message => "Please enter a genotype file",
			    -icon    => "warning",
			    -type    => "ok",
			    );
	return;
    }
    
    if (! $input_anno ) {
	$mw -> messageBox (-title   => "Input file missing",
			   -message => "Please enter an annotation file",
			   -icon    => "warning",
			   -type    => "ok",
			   );
	return;
    }
    if (! $output_name ) {	
	$mw -> messageBox (-title   => "Output file missing",
			   -message => "Please enter an output name",
			   -icon    => "warning",
			   -type    => "ok", 
			   );
	return;
    }

    # Test the validity of Illumina specific input and produce error messages:
    if ( $provider eq "Illumina" ) {
       	if (! $input_map ) {
	    $mw -> messageBox (-title   => "Input file missing",
			       -message => "Please enter a map file",
			       -icon    => "warning",
			       -type    => "ok",
			       );
	    return;
	}
	
	if (! $input_matrix ) {
	    $mw -> messageBox (-title   => "Input file missing",
			       -message => "Please enter a population matrix file",
			       -icon    => "warning",
			       -type    => "ok",
			       );
	    return;
	}
    }

    
    # Lock the start button
    &Lock_buttons();
    $mw -> update();

    # Print summary of the job requested:
    $log -> insert( "end", 
		    "Job launched with the following parameters:\n".
		    "\tProvider        = $provider\n".
		    "\tChip            = $chip\n".
		    "\tPopulation      = $population\n".
		    "\tGenotype file   = $input_geno\n". 
                    "\tAnnotation file = $input_anno\n".
		    "\tOutput name     = $output_name\n" );
    if ( defined $input_skip) {
	$log -> insert ("end", 
			"\tSkip list       = $input_skip\n" );
    }
    if ($provider eq "Illumina") {
	$log -> insert( "end", 
			"\tMap file        = $input_map\n".
			"\tMatrix file     = $input_matrix\n", 
			"\tFormat          = $format\n" 
			);
    }
    $log -> insert( "end", "\n" );
    $log -> see( "end" );
    $mw -> update();

    &Launch_formatting();

        # Unlock the start button. 
    &Unlock_buttons();
    $log -> insert ("end", 
		    "You can now process another job. But  be carefull to change ".
		    "the output name unless you want your previous result to be overwritten.\n\n" );
    $log -> see( "end" );
    $mw -> update();
    
}

sub Launch_formatting {
    
    # Affymetrix
    if ( $provider eq "Affymetrix" ) {
	
	# GenomeWideSNP_6
	if ( ($GUI && $chip eq $affy_chips_ref -> [0] ) ||
	     ($batch && $chip_number == 1) ) {
	    &Format_Affy_1; }
	
	# Other chips:
	else {
	    die "You have not entered a chip provider and a chip, "
		."or the chip you requested is not supported yet.\n";}
    }
    
    # Illumina
    if ( $provider eq "Illumina" ) {
	
	#quad410
	if ( ( $GUI && $chip eq $illu_chips_ref -> [0] ) ||
	     ( $batch && $chip_number == 1) ){
	    &Format_Illu_1; 
	}
	
	# Other chips:
	else {
	    die "You have not entered a chip provider and a chip, "
	       ."or the chip you requested is not supported yet.\n";}
    }
}


################################################################
###                        AFFYMETRIX                        ###
################################################################

sub Format_Affy_1 {

    # Variable declarations
    my $i           = 1;
    my $file_length = undef;
    my $progress    = 0;
    my %allele_hash;
    my %ID_hash;
    my ($AB, $A, $B);
    my $snp;
    my $chr;
    my $tmp;
    my $header_line = TRUE;
    my $start_of_header;
    my $frequency;
    my $genetic_position;
    my $unfound_frequency   = 0;
    my $unfound_position    = 0;
    my $number_of_SNPs      = 0;
    my @previous_line;
    
    # Open IO handles
    &Open_Output_Handles;
    &Open_Input_Handles;
    
    #Print some header:
    print $OUTPUT_FREQ "SNP_ID A_Frequency\n";
    print $OUTPUT_KEY "SNP_ID\tA_allele\tB_allele\n";
    

    ######                                  ######
    ### Parse the Affymetrix annotation file   ###
    ######                                  ######
    
    # Initialize the variables that guide the reading of the file
    $i           = 1;
    $header_line = TRUE;
    $progress    = 0;

    # Create a progress messages:
    if ( $GUI) {
	$log -> insert( "end", "Getting info on the annotation file:\n\t$input_anno.\n\n" );
	&My_Update() or return;
	$file_length = &CountLines($input_anno);
	$log -> insert( "end", "Reading the annotation file: $progress% done.\n");
	$log -> see( "end" );
	$integrity -> insert( "end", "For the annotation file $input_anno, Chip2Spell reads $file_length lines " );
	&My_Update() or return;
    }
    

    while ( <$INPUT_ANNO> ) {
	
	# Get a line of tab separated values input and creates an array of values
	chomp;
	s/\r//; 		# remove dos carriage return if they are present
	my @line = split /\t/, $_;
	
	# Skip comment lines
	if ( substr($line[0], 0, 1) eq "#" ) {
	    $i++;
	    next;
	}

	# Skip header line
	if ( $header_line ) {
	    #save the first fields of the header line for the integrity check.
	    $start_of_header = "$line[0]||$line[1]||$line[2]||$line[3]||$line[4]";
	    $header_line = FALSE;
	    $i++;
	    next;
	}
	
	# Count the processed SNPs:
	$number_of_SNPs ++;

	#Send output to the 'key' file:
	print $OUTPUT_KEY "$line[0]\t$line[9]\t$line[10]\n";
	
	# Create a hash with the allele A and allele B values 
	# and with the AffyProbeID as the key
	$allele_hash{ $line[0] } = $line[9] . $line[10];
	
	# Create a hash with rs numbers and probe ID.
	# and with the rs number as a key.
	$ID_hash{$line[2]} = $line[0];
	

	# Rename the chromosomes using numbers only.
	$chr = $line[3];
	if    ( $chr =~ /xy/i ) { $chr = 25; }
	elsif ( $chr =~ /mt/i ) { $chr = 26; }
	elsif ( $chr =~ /x/i  ) { $chr = 23; }
	elsif ( $chr =~ /y/i  ) { $chr = 24; }

	

	# Send genetic map information to a temporary file to be sorted later
	# The regex will capture the genetic position calculated from deCode.
	# If the regex does not capture, print "---"
	if ($line[12] =~ /^\s*(\d+\.\d+)\s*\/\//) {
	    $genetic_position = $1;
	    print $TMP_WRITING "$chr\t$line[0]\t$line[4]\t$genetic_position\n";
	}
	else {
	    print $TMP_WRITING "$chr\t$line[0]\t$line[4]\t---\n";
	    $unfound_position ++;
	}
	
	# Send output to allele frequency file:
	# The regex will capture the frequency of Allele A for the given population
	# The regex looks for a number, followed by "//" followed by another number and "//" followed by 
	# the target population. It captures the initial number.
	# If the regex does not capture, print ---
	$line[15];
	if ($line[15] =~ /(\d+(\.\d+)?)\s*\/\/\s*\d+(\.\d+)?\s*\/\/\s*$population/i) {
	    $frequency = $1;

	    # The frequency is adjusted to remove the extremes values.
	    if ($frequency == 1) {$frequency = 0.99;}
	    if ($frequency == 0) {$frequency = 0.01;}

	    print $OUTPUT_FREQ "$line[0] $frequency\n";}

	else {
	    $unfound_frequency ++;}
	

	# Print integrity message.
	if ($i == 100 && $GUI ) {
	    my $number_of_columns = @line;
	    $genetic_position = substr($genetic_position, 0, 6) . "..." ;
	    
   
# Integrity "heredoc" message
	    my $integrity_message = <<MESSAGE;
and $number_of_columns columns (the number of columns is taken at the 100th line).\n
The first fields of the header line read by Chip2Spell are: $start_of_header.\n
For the 100th line, Chip2Spell makes the following readings:
  \"$line[0]\" as the probe set ID (it should look like \"SNP_A-123456\").
  \"$line[2]\" as the rs number (it should look like. \"rs123456\").
  \"$line[3]\" as the chromosome (it should look like \"4\" or \"Y\").
  \"$line[4]\" as the physical position (it should look like \"123456\").
  \"$line[5]\" as the strand (it should look like \"+\" or \"-\").
  \"$line[9]\" as the A allele (it should look like \"A\" or \"G\").
  \"$genetic_position\" as the genetic position (it should look like \"123.456...\").
  \"$frequency\" as the allele frequencies (it should look like \"0.123\").\n
MESSAGE

	    $integrity -> insert( "end", $integrity_message );
	    $integrity -> see( "end" );
	    &My_Update() or return;
	}

	# Update main window every 1000 lines 
	if (! ($i % 1000) && $GUI ) { &My_Update() or return; }

	# Print progress messages every 1%.
	$i++;
	if ( $GUI ) {
	    if (! ($i % ($file_length/100))) {
		$progress = ((100*$i)-((100*$i)%$file_length))/$file_length;
		$log -> delete ("end -2 lines", "end");
		$log -> insert( "end", "\nReading the annotation file: $progress% done.\n" );
		&My_Update() or return;
	    }
	}
    } # End of "while ( <$INPUT_ANNO. )
  

    if ( $GUI ) {

	# Progress message:
	$log -> delete( "end -2 lines", "end");
	$log -> insert( "end", "\nReading the annotation file: 100% done.\n\n" );
	
	# Integrity message:
	$integrity -> insert( "end", "Information was read for $number_of_SNPs SNPs. The genetic position was not found in $unfound_position cases; and the A allele frequency was not found in $unfound_frequency cases.\n\n" );
	# Progress message: 
	$log -> insert( "end", "Sorting the map file:\n\t$output_name.map.\nThis might take a minute or two.\n\n");
	$log -> see( "end" );
	$integrity -> see( "end" );
	&My_Update() or return;
    }
    
    &Sort_Map;
    
 
    ######                                ######
    ### Reading the Affymetrix genotype file ### 
    ######                                ######
    
    # Initialize the variables that guide the reading of the file
    $i           = 1;
    $progress    = 0;
    $header_line = TRUE;
    
    # Create progress message:
    if ( $GUI ) {
	$log -> insert( "end", "Getting info on the genotype file:\n\t$input_geno.\n\n");
	&My_Update() or return;
	$file_length = &CountLines($input_geno);	
	$log -> insert( "end", "Reading the genotype file: $progress% done.\n" );
	$integrity -> insert( "end", "For the genotype file $input_geno, Chip2Spell reads $file_length lines " );
	&My_Update() or return;
    }
    
    while(<$INPUT_GENO>) {
	
	# Take a line of tab separated values input and turn it into an array of values
	chomp;
	s/\r//;
	my @line = split /\t/, $_;
	
	if ( $header_line ) {

	    # Search the first line containing genotype calls:
	    if ( ($line[1] =~ m/^([ACGT]{2}|0)$/i)  &&  ($line[2] =~ m/^([ACGT]{2}|0)$/i)  && $line[$#line] =~ m/^([ACGT]{2}|0)$/i ) {
		
		# Then the previous line is the header.
		$start_of_header = "$previous_line[0]||$previous_line[1]||$previous_line[2]||$previous_line[3]";

		# Create appropriate header for Alohomora:
		# The title "Marker" is replaced by "SNP ID"
		# Underscores in sample names are removed and "_Call" is added
		foreach my $field (@previous_line) {
		    
		    if ($field eq "Marker") {$field = "SNP ID";}
		    else {
			$field =~ s/_/-/g;
			$field = $field . "_Call";
		    }
		    print $OUTPUT_GENO "$field\t";
		}
		print $OUTPUT_GENO "\n";		
		
		$header_line = FALSE;
		$i++;
		next;
	    }
	    
	    else {
		@previous_line = @line;
		next;
	    }
	}
	
	# For each SNP line, the "0" must be replaced by "NC" and
	# the ACGT must be replaced by appropriate A or B.
	$snp = shift(@line);
	if ( exists $ID_hash{$snp}) {$snp = $ID_hash{$snp};}
	print $OUTPUT_GENO "$snp\t";
	
	# Extract the values of A and B alleles from the allele_hash
	$AB = $allele_hash{$snp};
	$A = substr($AB, 0, 1);
	$B = substr($AB, 1, 1);
	
	foreach my $field (@line) { 
	    if (!$field) {$tmp = "NC";}
	    else {
		
		# Transform ACGT format in AB format
		if (substr($field,0,1) eq $A){
		    $tmp = "A";}
		elsif (substr($field,0,1) eq $B) {
		    $tmp = "B";}
		if (substr($field,1,1) eq $A){
		    $tmp .= "A";}
		elsif (substr($field,1,1) eq $B){
		    $tmp .= "B";}
	    }

	    print $OUTPUT_GENO "$tmp\t";
	    
	}
	print $OUTPUT_GENO "\n";
	
	# Print data read at the 100th line for integrity check
	if ( $GUI && $i == 100 ) {
	    my $number_of_columns = @line;
	    
	    my $integrity_message = <<MESSAGE;
and $number_of_columns columns (the number of columns is taken at the 100th line).\n
The first five fields of the header line read by Chip2Spell are: $start_of_header.\n
For the 100th line, Chip2Spell makes the following readings:
  \"$snp\" as the SNP name (it should look like \"rs123456\" or \"SNP_A-123456\").
  \"$line[0]\" and \"$line[$#line]\" as the first and last genotype calls (it should look like \"AG\" or \"0\").\n
MESSAGE

            $integrity -> insert( "end", $integrity_message );
	    $integrity -> see( "end" );
	    &My_Update() or return;
	}

	# Update main window every 1000 lines 
	if (! ($1 % 1000) && $GUI ) { &My_Update() or return; }

	# Print progress message every 1%.
	$i++;
	if ( $GUI ) {
	    if (! ($i % ($file_length/100))) {
		$progress = ((100*$i)-((100*$i)%$file_length))/$file_length;
		$log -> delete("end - 2 lines", "end");
		$log -> insert( "end", "\nReading the genotype file: $progress% done.\n" );
		&My_Update() or return;
	    }
	}
    } # End of "while(<$INPUT_GENO>)"
    
    if ( $GUI ) {
	# Progress message
	$log -> delete( "end - 2 lines", "end");
	$log -> insert( "end", "\nReading the genotype file: 100% done.\n\n" );
	$log -> insert( "end", "Job done!\n\n" );
	$log -> see( "end" );
	$mw  -> update();
    }
    &Close_Output_Handles;
    &Close_Input_Handles;
    
} #end of sub Format_Affy_1


################################################################
###                        ILLUMINA                          ###
################################################################

sub Format_Illu_1 {

    # Variable declarations
    my %allele_hash;
    my $i               = 1;
    my $file_length     = undef;
    my $tmp;
    my $alleleA;
    my $alleleTotal;
    my $freqA;
    my $progress        = 0;
    my @sample_array;
    my $founders_list;
    my $chr;
    my $snp;
    my $after_data      = 0; 	# Refers to Illumina tag "[Data]" or "[Assay]"..
    my $before_controls = 1;  	# Refers to Illumina tag "[Controls]".ading a line in the file before the beginning
    my $header_line;
    my $A_T;			# Allele in ACGT-Top format that correspond to the A allele in AB format.
    my $B_T;			# Allele in ACGT-Top format that correspond to the B allele in AB format.
    my $A_F;			# Allele in ACGT-Forward format that correspond to the A allele in AB format.
    my $B_F;			# Allele in ACGT-Forward format that correspond to the B allele in AB format.
    my ($A, $B);
    my $start_of_header;
    
    # Open IO handles
    &Open_Output_Handles;
    &Open_Input_Handles;

    
    # Print Headers
    print $OUTPUT_FREQ "SNP_ID A_Frequency\n";
    print $OUTPUT_KEY "SNP_ID\tA_allele(FWD)\tB_allele(FWD)\tA_allele(TOP)\tB_allele(TOP)\n";
    
    ######                             ######
    ### Parse the Illumina input map file ###
    ######                             ######
    
    # Initialize the variables that guide the reading of the file
    $header_line = TRUE;
    $progress    = 0;
    $i           = 1;

    #Creating progress message:
    if ( $GUI ) {
	$log -> insert( "end", "Getting info on the input map file:\n\t$input_map.\n\n");
	&My_Update or return;
       	$file_length = &CountLines( $input_map );
	$log -> insert( "end", "Reading the input map file: $progress% done.\n" );
	$integrity -> insert( "end", "For the map file $input_map, Chip2Spell reads $file_length lines " );
	$log -> see( "end" );
	$integrity -> see( "end" );
	&My_Update or return;
    }
    
    
     
    while ( <$INPUT_MAP> ) {

	# Take a line of tab separated values input and makes an array of values
	chomp;
	s/\r//;
	my @line = split /\t/, $_;	
	
	#Skip the header line
	if ($header_line){
	    
	    # Save the first fields of header for integrity check:
	    $start_of_header = "$line[0]||$line[1]||$line[2]||$line[3]";

	    $header_line = FALSE;
	    $i++;
	    next;
	}
	
	# Rename the chromosomes using numbers only.
	$chr = $line[1];
	if ( (lc($chr)) eq "x"  ) { $chr = 23; }
	if ( (lc($chr)) eq "y"  ) { $chr = 24; }
	if ( (lc($chr)) eq "xy" ) { $chr = 25; }
	if ( (lc($chr)) eq "mt" ) { $chr = 26; }
	
	#Send output to a temporary map file.
	print $TMP_WRITING "$chr\t$line[0]\t$line[3]\t$line[2]\n";
	
	# Print integrity message for the 100th line of Illu map file.
	if ( $GUI && $i == 100 ) {
	    my $number_of_columns = @line;

	    # Integrity here-doc:
	    my $integrity_message = <<MESSAGE;
and $number_of_columns columns (the number of columns is taken at the 100th line).\n
The first fields of the header line read by Chip2Spell are: $start_of_header.\n
For the 100th line, Chip2Spell makes the following readings:
  \"$line[0]\" as the SNP name (it should look like \"123456\" or \"rs123456\").
  \"$line[1]\" as the chromosome (it should look like \"4\" or \"MT\").
  \"$line[2]\" as the physical position (it should look like \"123456\").
  \"$line[3]\" as the genetic position (it should look like \'123.456\').\n
MESSAGE
            $integrity -> insert( "end", $integrity_message );
	    $integrity -> see( "end" );
	    &My_Update() or return;
	}	    

	# Update main window every 1000 lines 
	if (! ($1 % 1000) && $GUI ) { &My_Update() or return; }

	# Print a progress messages every 1%.
	$i++;
	if ( $GUI) {
	    if (! ($i % ($file_length/100))) {
		$progress = ((100*$i)-((100*$i)%$file_length))/$file_length;
		$log -> delete ("end - 2 lines", "end" );
		$log -> insert ("end", "\nReading the input map file: $progress% done.\n" );
		&My_Update or return;
	    }
	}
    } # end of "while(<$INPUT_MAP>)
    
    if ( $GUI ) {
	$log -> delete ("end - 2 lines", "end");
	$log -> insert ("end", "\nReading the input map file: 100% done.\n\n");
	&My_Update() or return;
       
	# Sort the file output.map
	$log -> insert( "end", "Sorting the map file:\n\t$output_name.map.\nThis might take a minute or two.\n\n");
	$log -> see( "end" );
	&My_Update() or return;
    }


    &Sort_Map;
    
    
    
    ######                              ######
    ### Parse the illumina annotation file ###
    ######                              ######
    

    # Reinitialize the variables that guide the reading of the file
    $after_data      = FALSE; 
    $before_controls = TRUE;	
    $header_line     = TRUE;
    $i               = 1;
    $progress        = 0;
    
    # Create progress message:
    if ( $GUI ) {
	$log -> insert( "end", "Getting info ont the annotation file:\n\t$input_anno.\n\n" );
	&My_Update() or return;

	$file_length = &CountLines( $input_anno );
	
	$integrity -> insert( "end", "For the annotation file $input_anno, Chip2Spell reads $file_length lines " );
	$log -> insert( "end", "Reading the annotation file: $progress% done.\n" );
	$log -> see( "end" );
	$integrity -> see( "end" );
	&My_Update() or return;
    }
    
    while(<$INPUT_ANNO>) {
	
	# Take a line of tab separated values input and make an array of values
	chomp;
	s/\r//; 
	my @line = split /\t/, $_;
	
	# Skip the information lines
	if (! $after_data) {
	    if ( $line[0] =~ /\[Assay\]/i ) { $after_data = TRUE; }
	    $i++;
	    next;
	}
	
	if ( $before_controls ) {
	    
	    # Skip the header line:
	    if ($header_line) {
		
		# Save the first fields of the header for the integrity check
		$start_of_header = "$line[0]||$line[1]||$line[2]||$line[3]";
		
		$header_line = FALSE;
		$i++;
		next;
	    }
	    
	    # Skip the "[Controls]" line and the following lines.
	    if ( $line[0] =~ /\[Controls\]/i ) {
		$before_controls = FALSE;
		$i++;
		next;
	    }
	    
	    # Gets the SNP alleles using match
	    if (!($line[3] =~ m/\[([ACGT])\/([ACGT])\]/)){
		warn "for SNP $line[0], problem finding A and B in $line[3]. \$1 = $1 and \$2 = $2.\n";}
	    
	    # The values of the SNP are first assigned to each allele, 
	    # but those values will be modified according to actual B/T F/R data.
	    $A_T = $1; 
	    $A_F = $1;
	    $B_T = $2;
	    $B_F = $2;
	    
	    # Gets the Bottom/Top, Forward/Reverse information from the Illumina
	    # name using a regex match.
	    if (! ($line[0] =~ m/_(T|B)_(F|R)_/)) {
		warn "Problem while searching for _B_F_ at $line[0]\n";}
	     
	    # If Illm is top and correspond to forward, Top and Forward correspond 
	    # to the SNP, so we have nothing to change.	}
	    
	    # If Illm is Top and correspond to Reverse. Top = SNP 
	    # and Forward = transliteration of the SNP.
	    if ($1 eq "T" && $2 eq "R") {
		$A_F =~ tr/ACGT/TGCA/;
		$B_F =~ tr/ACGT/TGCA/;
	    }
	    
	    # If Illm is Bot and  Forward, Top is tr=~SNP and Forward = SNP
	    elsif ($1 eq "B" && $2 eq "F") {
		$A_T =~ tr/ACGT/TGCA/;
		$B_T =~ tr/ACGT/TGCA/;
	    }
	    
	    # If Illm is Bot and R, Top =~ trSNP and Forward =~ tr SNP.
	    elsif ($1 eq "B" && $2 eq "R") {
		$A_T =~ tr/ACGT/TGCA/;
		$B_T =~ tr/ACGT/TGCA/;
		$A_F =~ tr/ACGT/TGCA/;
		$B_F =~ tr/ACGT/TGCA/;
	    }
	    
	    # Send A and B alleles to OUTPUT_KEY
	    print $OUTPUT_KEY "$line[1]\t$A_F\t$B_F\t$A_T\t$B_T\n";
	    
	    # Create a hash with the allele A and B values.
	    $allele_hash{$line[1]} = $A_F . $B_F . $A_T . $B_T;
	 
	    # Print integrity message for the 100th line:
	    if ( $GUI && $i == 100 ) {
		my $number_of_columns = @line;
		
		# Integrity "here-doc" message:
		my $integrity_message = <<MESSAGE;
and $number_of_columns columns (the number of columns is taken at the 100th line).\n
The first fields of the header line read by Chip2Spell are $start_of_header.\n
For the 100th line, Chip2Spell makes the following readings:
  \"$line[0]\" as the IlmnID (it should look like \"rs123456_B_R_123456\").
  \"$line[1]\" as the name of the SNP (it should look like \"rs123456\").
  \"$line[3]\" as the polymorphism (it should look like \"[A/G]\").\n\n
MESSAGE
                $integrity -> insert( "end", $integrity_message );
		$integrity -> see( "end" );
		&My_Update() or return;
	    }

	    # Update main window every 1000 lines 
	    if (! ($1 % 1000) && $GUI ) { &My_Update() or return; }
    
	    # Print progress message every 1%.
	    $i++;
	    if ( $GUI ) {
		if (! ($i % ($file_length/100))) {
		    $progress = ((100*$i)-((100*$i)%$file_length))/$file_length;
		    $log -> delete ("end - 2 lines", "end");
		    $log -> insert ("end", "\nReading the annotation file: $progress% done.\n" );
		    &My_Update() or return;
		}
	    }
	}
    } # End of while($INPUT_ANNO) 

    # Progress message:
        if ( $GUI ) {
	$log -> delete ("end - 2 lines", "end");
	$log -> insert ("end", "\nReading the annotation file: 100% done.\n\n");
	$log -> see( "end" );
	&My_Update() or return;
    }
    
    ######                                 ######
    ### Parse illumina population matrix file ###
    ######                                 ######
    
    # Reinitialize the variables that guide the reading of the file
    $after_data  = FALSE;
    $header_line = TRUE;
    $progress    = 0;
    $i           = 1;

    # Create progress message:
    if ( $GUI ) {
	$log -> insert ("end", "Getting info on the population matrix file:\n\t$input_matrix.\n\n");
	&My_Update() or return;

	$file_length = &CountLines( $input_matrix );

	$integrity -> insert( "end", "For the population matrix file $input_matrix, Chip2Spell reads $file_length lines " );
	$log -> insert( "end", "Reading the population matrix file: $progress% done.\n");
	$integrity -> see( "end" );
	$log -> see( "end" );
	&My_Update() or return;
    }
    
    
    while(<$INPUT_MATRIX>) {

	# Take a line of tab separated values input and make an array of values
	chomp;
	s/\r//;
	my @line = split /\t/, $_;
	
	# Skip the information lines
	if ( not $after_data ) {
	    if ($line[0] =~ /\[Data\]/i) {$after_data = TRUE;}
	    $i++;
	    next;
	}
	
	if ( $header_line ) {

	    # Save first fields of the header line for integrity check:
	    $start_of_header = "$line[0]||$line[1]||$line[2]||$line[3]";
	    
	    # Choose founders list for the correct population
	    if ($population eq "Caucasian") {
		$founders_list = $founders_list_CEU;}
	    elsif ($population eq "Han Chinese") {
		$founders_list = $founders_list_CHB;}
	    elsif ($population eq "Yoruban") {
		$founders_list = $founders_list_YRI;}
	    elsif ($population eq "Japanese") {
		$founders_list = $founders_list_JPT;}
	    
	    
	    # Identify the columns that correspond to founders
	    foreach my $column (@line) {
		if ($founders_list =~ /$column/) {
		    $column = "true";}
		else {
		    $column = undef;
		}
	    }
	    
	    # Store the result in a new array.
	    @sample_array = @line;
	    
	    # The first column was empty, so it was an automatic match, but we 
	    # still want it to be false. 
	    $sample_array[0] = undef;
	    
	    # Count the totalAllele number.
	    for my $sample (@sample_array) {
		if ($sample) {$alleleTotal += 2;}
	    }
	    
	    $header_line = FALSE;
	    $i++;
	    next;   
	}	
	
	# The following is done for each SNP lines of the population matrix file
	
	$alleleA = 0;		# running count of A alleles

	# Find the A allele for a given SNP
	$A = substr($allele_hash{$line[0]}, 0, 1);
	
	for (my $j = 1; $j<= $#sample_array; $j++) {
	    if ($sample_array[$j]) {
		if ( substr($line[$j], 0, 1 ) eq $A) {
		    $alleleA ++;}
		if ( substr($line[$j], 1, 1 ) eq $A) {
		    $alleleA ++;}
	    }
	}
	$freqA = $alleleA / $alleleTotal;

	# Adjust extreme values
	if ($freqA == 0) {$freqA = 0.01;}
	if ($freqA == 1) {$freqA = 0.99;}
	
	print $OUTPUT_FREQ "$line[0] $freqA\n";
	
	# Print integrity message for the 100th line:
	if ( $GUI && $i == 100 ) {
	    my $number_of_columns = @line;

	    # Integrity "here-doc" message
	    my $integrity_message = <<MESSAGE;
and $number_of_columns columns (the number of columns is taken at the 100th line).\n
The first five fields of the header line read by Chip2Spell are: $start_of_header.\n
For the 100th line, Chip2Spell makes the following readings:
  \"$line[0]\" as the SNP name (it should look like \"123456\" or \"rs123456\").
  \"$line[1]\" and \"$line[$#line]\" as the first and last genotype call (it should look like \"AG\" or \"TC\").\n
MESSAGE
            $integrity -> insert( "end", $integrity_message );
	    $integrity -> see( "end" );
	    &My_Update() or return;
	}

	# Update main window every 1000 lines 
	if (! ($1 % 1000) && $GUI ) { &My_Update() or return; }
    
	# Print progress message every 1%.
	$i++;
	if ( $GUI ) {
	    if (! ($i % ($file_length/100))) {
		$progress = ((100*$i)-((100*$i)%$file_length))/$file_length;
		$log -> delete( "end - 2 lines", "end");
		$log -> insert( "end", "\nReading the population matrix file: $progress% done.\n" );
		&My_Update() or return; 
	    }
	}
    } # End of "while(<$INPUT_MATRIX>)
    
    if ( $GUI ) {
	$log -> delete( "end - 2 lines", "end");
	$log -> insert( "end", "\nReading the population matrix file: 100% done.\n\n" );
	&My_Update() or return;
    }
    
    
    ######                            ######
    ### Parse the illumina genotype file ###
    ######                            ######
    
    # Reinitialize the variables that guide reading the file
    $after_data  = FALSE;
    $header_line = TRUE;
    $progress    = 0;
    $i           = 1;

    # Create progress messages:
    if ( $GUI ) {
	$log -> insert( "end", "Getting info on the genotype file:\n\t$input_geno.\n\n" );
	&My_Update() or return;

	$file_length = &CountLines($input_geno);

	$log -> insert( "end", "Reading the genotype file: $progress done.\n");
	$integrity -> insert( "end", "For the genotype file $input_geno, Chip2Spell reads $file_length lines " );
	$integrity -> see( "end" );
	$log -> see( "end" );
	&My_Update() or return;
    }
    
    while (<$INPUT_GENO>) {
	
	# Take a line of tab separated values input and make an array of values
	chomp;
	s/\r//;
	my @line = split/\t/, $_;
	
	# Skip the information lines
	if (! $after_data) {
	    if ( $line[0] =~ /\[Data\]/i ) {$after_data = TRUE;}
	    $i++;
	    next;
	}
	
	# Create the header line
	if ($header_line) {
	    
	    # Save the first fields of the header line for integrity check.
	    $start_of_header = "$line[0]||$line[1]||$line[2]||$line[3]";

	    # Print SNP_ID
	    shift @line;
	    print $OUTPUT_GENO "SNP ID\t";
	    
	    foreach my $field (@line) {

		# Create appropriate header for Alohomora:
		# Underscores in sample names are removed and "_Call" is added
		$field =~ s/_/-/g;
		$field = $field . "_Call";
		print $OUTPUT_GENO "$field\t";
	    }

	    print $OUTPUT_GENO "\n";
	    
	    $header_line = FALSE;
	    $i++;
	    next;
	}
	

	if ( $format eq "AB" ) {

	    # Print SNP_ID 
	    $snp = shift (@line);
	    print $OUTPUT_GENO "$snp\t";
	    
	    foreach my $field (@line) {
	
		# Adjust the format:
		# Replace "--" by "NC" and send to output 
		if (substr($field, 0, 2)  eq "--") {
		    $tmp = "NC";}
		else {
		    $tmp = substr($field, 0, 2);}
	
		print $OUTPUT_GENO "$tmp\t";
	    }
	    print $OUTPUT_GENO "\n";
	}
	
	# If the original format is "ACGT-top", transform to "AB", 
	# replace "--" by "NC" and send to output.
	elsif ($format eq "ACGT-FWD") {
	    
	    # Print SNP_ID
	    $snp = shift (@line);
	    print $OUTPUT_GENO "$snp\t";
	    
	    #allele_hash: first letter is A allele in Top format, and second letter is B allele in top format
	    $A = substr($allele_hash{$snp},0,1);
	    $B = substr($allele_hash{$snp},1,1);
	    
	    foreach my $field (@line) {

		# Adjust the format: 
		# Replace "--" by "NC" and send to output.
		if (substr ($field, 0,2) eq "--") {
		    $tmp = "NC";}

		# Adjust the format:
		# Replace an ACGT call by an AB call
		else {
		    if (substr ($field, 0, 1) eq $A && substr($field,1,1) eq $A) {
			$tmp = "AA";}
		    elsif (substr ($field, 0, 1) eq $B && substr($field, 1,1) eq $B) {
			$tmp = "BB";}
		    elsif (substr($field, 0, 1) eq $A && substr($field,1,1) eq $B) {
			$tmp = "AB";}
		    elsif (substr($field, 0, 1) eq $B && substr($field,1,1) eq $A) {
			$tmp = "AB";}
		}
		print $OUTPUT_GENO "$tmp\t";
	    }
	    print $OUTPUT_GENO "\n";
	}
	# If the original format is "ACGT-top", transform to "AB", 
	# replace "--" by "NC" and send to output.
	elsif ($format eq "ACGT-TOP") {
	    
	    # Print SNP_ID
	    $snp = shift (@line);
	    print $OUTPUT_GENO "$snp\t";
	    

	    # The third letter of allele_hash is the A allele in forward format, 
	    # and the last is the B allele in fwd format.
	    $A = substr($allele_hash{$snp},2,1);
	    $B = substr($allele_hash{$snp},3,1);
	    
	    foreach my $field (@line) {

		# Adjust the format: 
		# Replace "--" by "NC" and send to output.
		if (substr ($field, 0,2) eq "--") {
		    $tmp = "NC";}

		# Adjust the format:
		# Replace an ACGT call by an AB call
		else {
		    if (substr ($field, 0, 1) eq $A && substr($field,1,1) eq $A) {
			$tmp = "AA";}
		    elsif (substr ($field, 0, 1) eq $B && substr($field, 1,1) eq $B) {
			$tmp = "BB";}
		    elsif (substr($field, 0, 1) eq $A && substr($field,1,1) eq $B) {
			$tmp = "AB";}
		    elsif (substr($field, 0, 1) eq $B && substr($field,1,1) eq $A) {
			$tmp = "AB";}
		}
		print $OUTPUT_GENO "$tmp\t";
	    }
	    print $OUTPUT_GENO "\n";
	}
	
	
	# Print integrity check message.
	if ( $GUI && $i == 100 ) {
	    my $number_of_columns = @line;
	    
# Integrity "here-doc" message:
	    my $integrity_message = <<MESSAGE;
and $number_of_columns columns (the number of columns is taken at the 100th line).\n
The first fields of the header line read by Chip2Spell are: $start_of_header.\n
For the 100th line, Chip2Spell makes the following readings:
  \"$snp\" as the SNP name (it should look like \"rs123456\").
  \"$line[0]\" and \"$line[$#line]\" as the first and last genotype call (it should look like \"CG\" or \"AB\" or \"--\"),
MESSAGE
            $integrity -> insert( "end", $integrity_message );
	    $integrity -> see( "end" );
	    &My_Update() or return;
	}
	    

	# Update main window every 1000 lines 
	if (! ($1 % 1000) && $GUI ) { &My_Update() or return; }
	
	# Print progress message every 1%
	$i++;
	if ( $GUI ) {
	    if (! ($i % ($file_length/100))) {
		$progress = ((100*$i)-((100*$i)%$file_length))/$file_length;
		$log -> delete( "end - 2 lines", "end");
		$log -> insert( "end", "\nReading the genotype file: $progress% done.\n" );
		&My_Update() or return;
	    }
	}
    } # End of while($<INPUT_GENO>)
    
    if ( $GUI ) {
	$log -> delete( "end - 2 lines", "end" );
	$log -> insert( "end", "\nReading the genotype file: 100% done.\n\n" );
	$log -> insert( "end", "Job done!\n\n" );
	$log -> see( "end" );
	&My_Update() or return;
    }
    
    &Close_Output_Handles;
    &Close_Input_Handles; 

} # End of Format_Illu_1    

sub Open_Output_Handles {

    if (! open $OUTPUT_GENO, ">", "$output_name.geno") {
	die "The file $output_name.geno could not be created: $!";}
    if (! open $OUTPUT_MAP, ">", "$output_name.map") {
	die "The file $output_name.anno could not be created: $!";}
    if (! open $OUTPUT_FREQ, ">", "$output_name.freq") {
	die "The file $output_name.freq could not be created: $!";}
    if (! open $OUTPUT_KEY, ">", "$output_name.key") {
	die "The file $output_name.key could not be created: $!";}
    if (! open $TMP_WRITING,">", "$output_name.tmp") {
	die "Could not open TMP_WRITING.\n";}
}

sub Close_Output_Handles {
    if (! close $OUTPUT_GENO) {
	die "The handle OUTPUT_GENO could not be closed: $!";}
    if (! close $OUTPUT_MAP) {
	die "The handle OUTPUT_MAP could not be closed: $!";}
    if (! close $OUTPUT_FREQ) {
	die "The handle OUTPUT_FREQ could not be closed: $!";}
    if (! close $OUTPUT_KEY) {
	die "The handle OUTPUT_KEY could not be closed: $!";}
    if (! close $TMP_WRITING) {
	die "Could not close TMP_WRITING\n";}
}

sub Open_Input_Handles {
    if (! open $INPUT_GENO, "<", "$input_geno") {
	die "The file $input_geno could not be opened: $!";}
    if (! open $INPUT_ANNO, "<", "$input_anno") {
	die "The file $input_anno could not be opened: $!";}

    # The following files are not always used, so they won't be associated
    # with a 'die' that stops the program each time the file is not used.
    if ($provider eq "Illumina") {
	if ( ! open $INPUT_MAP, "<", "$input_map" ) {
	    die "The file $input_map could not be opened: $!";}
	if (! open $INPUT_MATRIX, "<", "$input_matrix" ) {
	    die "The file $input_matrix could not be opened: $!";}
    }
}

sub Close_Input_Handles {
    
    # General files
    if (! close $INPUT_GENO) {
	die "The handle INPUT_GENO could not be closed: $!";}
    if (! close $INPUT_ANNO) {
	die "The handle INPUT_ANNO could not be closed: $!";}
    
    # Illumina files
    if ($provider eq "Illumina") {
	if (! close $INPUT_MAP ) {
	    die "The handle INPUT_MAP could not be closed: $!";}
	if (! close $INPUT_MATRIX ) {
	    die "The handle INPUT_MATRIX could not be closed: $!";}
    }
}

sub Sort_Map {
    if (! open our $TMP_READING, "<", "$output_name.tmp") {
	die "Could not open TMP_READING.\n";}
    
    print $OUTPUT_MAP "Chr\tSNP\tGenetic_position\tPhysical_position\n";
    my %cache1;
    my %cache2;

    ### Sort map numerically according to fields 0 and 2
    print $OUTPUT_MAP sort {

	( ($cache1{$a} ||= &Key1($a) ) <=> ($cache1{$b} ||= &Key1($b) ) )
	    ||
	( ($cache2{$a} ||= &Key2($a) ) <=> ($cache2{$b} ||= &Key2($b) ) )
 
	}(<$TMP_READING>);

    if (! close $TMP_READING) {
	die "Could not close TMP_READING filehandle.\n";}

#   if (! unlink ("$output_name.tmp")) {
#	die "Could not delete temporary map file \"$output_name.tmp\".\n";}
}

sub Key1 {
    # Updating the GUI
    if ($GUI) {
	$sort_counter++;
	if (not ( $sort_counter % 1000 )) {
	    $mw -> update();}
    }
    my @tmp = split/\t/, $_[0];
    return $tmp[0]; 
}
sub Key2 {
    my @tmp = split/\t/, $_[0];
    return $tmp[2];
}

# Remove leading and trailing whitespace from a string.
sub Trim {
    my $string = $_[0];
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
}

### Check if the input $_[0] can be read.
### If not, print an error message.
sub CheckReadMode {
    if (! -e $_[0]) {
	print "\"$_[0]\" does not exist.\n";}
    elsif (! -f $_[0]) {
	print "\"$_[0]\" is not a regular file.\n";}
    elsif (! -r $_[0]) {
	print "You do not have permission to read \"$_[0]\".\n";}
    else {
	return;}
}


### Check to see if the program can write to its assigned four output files.
### If not, print an error message.
sub CheckWriteMode {
    if (! -w "$_[0].geno" &&  -e "$_[0].geno") {
	print "You do not have persmission to write to \"$_[0].geno\".\n";}
    elsif (! -w "$_[0].key" && -e "$_[0].key") {
	print "You do not have persmission to write to \"$_[0].key\".\n";}
    elsif (! -w "$_[0].freq" && -e "$_[0].freq") {
	print "You do not have persmission to write to \"$_[0].freq\".\n";}
    elsif (! -w "$_[0].map" && -e "$_[0].map") {
	print "You do not have persmission to write to \"$_[0].map\".\n";}
    else {
	return;}
}


### Count the lines of a file
sub CountLines {
    my $COUNT_LINES;
    my $count = 0;
    
    if (! open $COUNT_LINES, "<", $_[0]) {
	die "Could not open the filehandle COUNT_LINES for the file $_[0].\n";}
    
    while(<$COUNT_LINES>) {
	$count++;
	
	#update the GUI every 5000 lines.
	if ( $count % 5000 == 0 ) { &My_Update() or return;}
    }
    
    if (! close $COUNT_LINES) {
	die "Could not close the filehandle COUNT_LINES for the file $_[0].\n";}
    
    return $count;
}

sub Lock_buttons {
    for my $item ( $geno_button,     $anno_button,   # $skip_list_button,
		   $map_button,      $matrix_button,   $ab_rb,
		   $acgt_top_rb,     $acgt_fwd_rb,     $chip_be,
		   $ceu_rb,          $hcn_rb,          $jpt_rb,
		   $yrb_rb,          $affy_rb,         $illu_rb,
		   $start_button,    $reset_button,    $output_button,
                  ) {
	$item -> configure ( -state => "disabled" );
    }

    $exit_button -> configure( -text => "Abort", -command => sub {$abort = TRUE;} , );
}

sub Unlock_buttons {

    for my $item ( $geno_button,     $anno_button,   # $skip_list_button,
		   $map_button,      $matrix_button,   $ab_rb,
		   $acgt_top_rb,     $acgt_fwd_rb,     $chip_be,
		   $ceu_rb,          $hcn_rb,          $jpt_rb,
		   $yrb_rb,          $affy_rb,         $illu_rb,
		   $start_button,    $reset_button,    $output_button,
		   ) {
	$item -> configure ( -state => "normal" );
    }

    $exit_button -> configure( -text => "Exit", -command => sub {exit;}, );
}

sub My_Update {
    $mw -> update();

    # Clean-up work
    if ($abort) {
	# Clean-up handles
	&Close_Input_Handles();
	&Close_Output_Handles();

	# Delete temporary file if it still exists.
	if (-e "$output_name.tmp") {
	    if (! unlink ("$output_name.tmp")) {
		warn "Could not delete temporary map file \"$output_name.tmp\".\n";}
	}
	
	# Print messages
	$log -> insert( "end", "Job aborted...\n\n" );
	$integrity -> insert( "end", "Job aborted...\n\n" );
	$log -> see( "end" );
	$integrity -> see( "end" );
	$mw -> update();

	&Unlock_buttons();
	return FALSE;
    }
    return TRUE;
}

sub Affy_provider {
    $chip_be -> configure( -choices => $affy_chips_ref ) ;
    for my $item ($map_button, $matrix_button, $ab_rb, $acgt_fwd_rb, $acgt_top_rb) {
	$item -> configure ( -state => "disabled" );
    }
}

sub Illu_provider {
    $chip_be -> configure( -choices => $illu_chips_ref );
    for my $item ($map_button, $matrix_button, $ab_rb, $acgt_fwd_rb, $acgt_top_rb) {
	$item -> configure ( -state => "normal" );
    }
}

__END__


=pod

=head1 NAME

Chip2Spell

=head1 SYNOPSIS

Chip2Spell helps you prepare your genotype data for the use of Alohomora by formatting your data and complementing it from publicly available annotation files.

=head1 BATCH MODE

There are currently 11 options for the batch mode: 

 --annotation     --matrix
 --batch          --output
 --chip           --population
 --format         --provider
 --genotype       --skiplist
 --map

=over 2

=item ANNOTATION

 Use this option to enter your annotation file.
 Usage: "--annotation=annoFile.txt".
 Default: no annotation file is provided by default.

=item BATCH

 Use this option to enter the batch mode.
 Usage: "--batch__ .
 Default: graphical user interface is used unless the batch option is specified.

=item CHIP

 Use this option to enter a chip number. Currently, only 1 chip was explicitely tested for both Affymetrix and Illumina.
 Usage: "--chip=1".
 Defautl: chip = 1.

=item FORMAT

 Use this option to enter the original format of the data (this option is specific to Illumina).
 Usage: "--format=AB" or "--format=ACGT-TOP" or "--format=ACGT-FWD".
 Default: format = AB.

=item GENOTYPE

 Use this option to enter your genotype file.
 Usage: "--genotype=genoFile.txt".
 Default: no genotype file is provided by default.

=item MAP

 Use this option to enter your map file (this option is specific to Illumina).
 Usage: "--map=mapFile.txt".
 Default: no map file is provided by default.

=item MATRIX

 Use this option to enter your population genotype matrix file (this option is specific to Illumina). 
 Usage: "--matrix=matrixCEU.txt".
 Default: no matrix file is provided by default.

=item OUTPUT

 Use this option to provide an output name. This name will be used to create four output files: name.geno, name.map, name.freq, name.key.
 Usage: "--output=GivenName".
 Default: the default output name is "output".

=item POPULATION

 Use this option to specify a number corresponding to the population from which the frequencies are to be taken. 1 = Caucasian, 2 = Han Chinese, 3 = Japanses, 4 = Yoruban.
 Usage: "--population=1".
 Default: population=1 (Caucasian/CEU).

=item PROVIDER

 Use this option to specify your chip provider (Affymetrix or Illumina).
 Usage: "--provider=Affymetrix" or "--provider=Illumina".
 Defaut: --provider=Affymetrix

=item SKIPLIST

 # NOTE # This option is not currently implemented
 Use this option to specify a skip list file. The SNPs in this list will be ignored when your output files will be generated.
 Usage: "--skiplist=skipListFile.txt"
 Default: no skip list is provided by default.

=back

=head1 NOTES ON FORMAT

For every input files, chip2spell expect that the file is in a I<tab separated value> format.
For every input files, chip2spell will expect that the file is a linux file (i.e. that the lines do not end with \r characters). It is a good precaution to use the command "dos2unix" on all the files before using them with chip2spell. 

=head2 AFFYMETRIX

For affymetrix chips, chip2spell expects 2 files:

=over 3

=item 1.

a genotype file in the NANUQ format;

=item 2.

an annotation file in the Affymetrix format.

=back

=head3 Assumptions for the Affymetrix  genotype file:

=over 3

=item *

The first 5 lines are comment. 

=item *

The 6th line is the header and contains the names of the samples.

=item *

The first column contains the identifier of the SNP, either its rs number (e.g. "rs100002") or the Affymetrix probe set ID (e.g. "SNP_A-1793631" or "AFFX-SNP_8685025).

=item *

Each subsequent column contains the calls for a given sample (e.g. "AG" or "0").

=back

=head3 Assumptions for the Affymetrix annotation file:

=over 3

=item *

Lines starting by # are comments.

=item *

The 1st column contains the probe set ID (e.g. "SNP_A-17p4641" or "AFFX-SNP_8685025).

=item *

The 3rd column contains the rs number (e.g. "rs10000002).

=item *

The 3th column contains the chr (e.g. "1" or "MT").

=item *

The 5th column contains the physical position (e.g. "84647761").

=item *

The 6th column contains the strand (e.g. "-" or "+").

=item *

The 10th and 11th columns contains alleles A and B (e.g. "C" or "G").

=item *

The 13th column contains the genetic map (e.g. "116.765222864103 // AFMA081XE9 // AFMB320YF1 // D1S2889 // D1S2766 /// 108.402547694204 // D1S2889 // D1S2766 // --- // --- /// 102.95735950781 // --- // --- // 586552 // 615456"). A regular expression is used to capture the first number of the line followed by two "//".  

=item *

The 16th column contains the allele frequencies (e.g."0.6 // 0.4 // Han Chinese /// 0.55 // 0.45 // Caucasian /// 0.6222 // 0.3778 // Japanese /// 0.5167 // 0.4833 // Yoruban"). A regular expression is used to capture the number followed by // + another number + // + the name of the target population. 

=back

=head2 ILLUMINA
	
For Illumina, chip2spell expects four files:

=over 3

=item 1.

a genotype file;

=item 2.

an annotation file;

=item 3.

a map file;

=item 4.

a genotype matrix file for a given population.

=back	



=head3 Assumptions for the genotype files:

=over 3

=item * 

A line "[Data]" signals the beginning of the data section.

=item *

The first line after "[Data]" is the header line with the sample names.

=item *

The first column is the SNP name (e.g. "rs1000004").

=item *

Each subsequent column correspond to a sample (e.g. "AB" or "--" for a missing call).

=back

=head3 Assumptions for the annotation file:	 	

=over 3

=item *

A line "[Assay]" signals the beginning of the data section.

=item *

A line "[Controls]" signals the end of the data section and the beginning of the control section.

=item *

The first column contains the IlmnID (e.g. "200003-0_B_R_1526882017").

=item *

The second column contains the name of the SNP (most often, its rs number, e.g. "rs100004").

=item *

The fourth column shows the SNP (ex: "[A/G]").

=back

=head3 Assumptions for the map file:

=over 3

=item *

There are no information lines.

=item *

The first line is a header.

=item *

The first column contains the SNP name, (e.g. "200003" or "rs1000004").

=item *

The second column contains the chromosome (e.g. "4" or "Y").

=item *

The third column contains the physical position (e.g. "7455806").

=item *

The fourth column contains the genetic position, e.g. "52.4893").

=back
	
=head3 Assumptions for the genotype matrix file:

=over 3
 
=item *

A line "[Data]" signals the beginning of the data section.

=item *

The first line after "[Data]" contains the names of the samples.

=item *

The first column contains the SNP names (e.g. "200003" or "rs1000004").

=item *

Each subsequent column correspond to a sample (e.g. "AG").

=item *

The samples are taken from thoses listed in hapMap. HapMap documentation was used to identify the founders of the different populations and only the founders are used to calculate the allele frequencies for that population. 

=back

=head1 NOTES ON OUTPUT:

=head2 output.geno

A "_Call" is systematically added to all sample names. Missing calls are marked "NC".

=head2 output.map
	
Named chromosomes are replaced by numbered chromosomes:"x" = 23; "y" = 24; "xy" = 25; "mt" = 26.
	
=head2	output.freq

=head2	output.key

=head1 WHAT TO DO WHEN THINGS GO WRONG

Chip2Spell's "Integrity check" window helps you determine if something is going wrong, but it will not tell you why this is so, or what to do to correct the situation. Here are some of the reasons that might prevent Chip2Spell from working properly and some of the solutions that you can try.

=over 3

=item *

Transform your files to a tab separated value format.

=item *

Transform your files to a unix format (use the dos2unix utility).

=item *

Check whether the expected tags (for example "[Data]" or [Assay]") are present in your files. If they are not present, put them back manually.

=item * 

Take a look at the "NOTES ON FORMAT" section for the file that is not read properly and examine whether the relevant pieces of information are still displayed in the columns in which Chip2Spell expects them. Maybe Affymetrix or Illumina have changed slightly their documentation by adding or removing a column. Then, manually removing the added column or manually adding an empty column should fix the problem.

=back 

=cut
